@@ -1,4 +1,4 @@
-# Copyright (C) 2002 Free Software Foundation, Inc.
+# Copyright (C) 2002, 2004, 2006, 2008 Free Software Foundation, Inc.
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -11,9 +11,12 @@
 # GNU General Public License for more details.
 
 # You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-# 02111-1307, USA.
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+###############################################################
+# The main copy of this file is in Automake's CVS repository. #
+# Updates should be sent to automake-patches@gnu.org.         #
+###############################################################
 
 package Automake::Channels;
 
@@ -40,10 +43,17 @@ Automake::Channels - support functions for error and warning management
   # Turn on all channels of type 'warning'.
   setup_channel_type 'warning', silent => 0;
 
+  # Redirect all channels to push messages on a Thread::Queue using
+  # the specified serialization key.
+  setup_channel_queue $queue, $key;
+
+  # Output a message pending in a Thread::Queue.
+  pop_channel_queue $queue;
+
   # Treat all warnings as errors.
   $warnings_are_errors = 1;
 
-  # Exit with the greater exist code encountered so far.
+  # Exit with the greatest exit code encountered so far.
   exit $exit_code;
 
 =head1 DESCRIPTION
@@ -71,6 +81,7 @@ use vars qw (@ISA @EXPORT %channels $me);
 	      &setup_channel &setup_channel_type
 	      &dup_channel_setup &drop_channel_setup
 	      &buffer_messages &flush_messages
+	      &setup_channel_queue &pop_channel_queue
 	      US_GLOBAL US_LOCAL
 	      UP_NONE UP_TEXT UP_LOC_TEXT);
 
@@ -108,7 +119,7 @@ $warnings_are_errors = 0;
 
 =item C<UP_NONE>, C<UP_TEXT>, C<UP_LOC_TEXT>
 
-Possible values for the C<uniq_part> options.  This select the part
+Possible values for the C<uniq_part> options.  This selects the part
 of the message that should be considered when filtering out duplicates.
 If C<UP_LOC_TEXT> is used, the location and the explanation message
 are used for filtering.  If C<UP_TEXT> is used, only the explanation
@@ -126,10 +137,10 @@ use constant UP_LOC_TEXT => 2;
 
 Possible values for the C<uniq_scope> options.
 Use C<US_GLOBAL> for error messages that should be printed only
-once in the run of the program, C<US_LOCAL> for message that
+once during the execution of the program, C<US_LOCAL> for message that
 should be printed only once per file.  (Actually, C<Channels> does not
-now when files are changed, it relies on you calling C<reset_local_duplicates>
-when this happens.)
+do this now when files are changed, it relies on you calling
+C<reset_local_duplicates> when this happens.)
 
 =cut
 
@@ -143,7 +154,7 @@ use constant US_GLOBAL => 1;
 
 Channels accept the options described below.  These options can be
 passed as a hash to the C<register_channel>, C<setup_channel>, and C<msg>
-functions.  The possible keys, with there default value are:
+functions.  The possible keys, with their default value are:
 
 =over
 
@@ -170,12 +181,20 @@ The file where the error should be output.
 Whether the channel should be silent.  Use this do disable a
 category of warning, for instance.
 
+=item C<ordered =E<gt> 1>
+
+Whether, with multi-threaded execution, the message should be queued
+for ordered output.
+
 =item C<uniq_part =E<gt> UP_LOC_TEXT>
 
 The part of the message subject to duplicate filtering.  See the
 documentation for the C<UP_NONE>, C<UP_TEXT>, and C<UP_LOC_TEXT>
 constants above.
 
+C<uniq_part> can also be set to an arbitrary string that will be used
+instead of the message when considering duplicates.
+
 =item C<uniq_scope =E<gt> US_LOCAL>
 
 The scope of duplicate filtering.  See the documentation for the
@@ -246,6 +265,9 @@ use vars qw (%_default_options %_global_duplicate_messages
    exit_code => 1,
    file => \*STDERR,
    silent => 0,
+   ordered => 1,
+   queue => 0,
+   queue_key => undef,
    uniq_scope => US_LOCAL,
    uniq_part => UP_LOC_TEXT,
    header => '',
@@ -317,6 +339,13 @@ sub _merge_options (\%%)
 	  confess "unknown option `$_'";
 	}
     }
+  if ($hash->{'ordered'})
+    {
+      confess "fatal messages cannot be ordered"
+	if $hash->{'type'} eq 'fatal';
+      confess "backtrace cannot be output on ordered messages"
+	if $hash->{'backtrace'};
+    }
 }
 
 =item C<register_channel ($name, [%options])>
@@ -349,7 +378,7 @@ sub exists_channel ($)
 =item C<channel_type ($name)>
 
 Returns the type of channel C<$name> if it has been registered.
-Returns The empty string otherwise.
+Returns the empty string otherwise.
 
 =cut
 
@@ -397,6 +426,63 @@ sub _format_message ($$%)
   return $msg;
 }
 
+# _enqueue ($QUEUE, $KEY, $UNIQ_SCOPE, $TO_FILTER, $MSG, $FILE)
+# ------------------------------------------------------------
+# Push message on a queue, to be processed by another thread.
+sub _enqueue ($$$$$$)
+{
+  my ($queue, $key, $uniq_scope, $to_filter, $msg, $file) = @_;
+  $queue->enqueue ($key, $msg, $to_filter, $uniq_scope);
+  confess "message queuing works only for STDERR"
+    if $file ne \*STDERR;
+}
+
+# _dequeue ($QUEUE)
+# -----------------
+# Pop a message from a queue, and print, similarly to how
+# _print_message would do it.  Return 0 if the queue is
+# empty.  Note that the key has already been dequeued.
+sub _dequeue ($)
+{
+  my ($queue) = @_;
+  my $msg = $queue->dequeue || return 0;
+  my $to_filter = $queue->dequeue;
+  my $uniq_scope = $queue->dequeue;
+  my $file = \*STDERR;
+
+  if ($to_filter ne '')
+    {
+      # Do we want local or global uniqueness?
+      my $dups;
+      if ($uniq_scope == US_LOCAL)
+	{
+	  $dups = \%_local_duplicate_messages;
+	}
+      elsif ($uniq_scope == US_GLOBAL)
+	{
+	  $dups = \%_global_duplicate_messages;
+	}
+      else
+	{
+	  confess "unknown value for uniq_scope: " . $uniq_scope;
+	}
+
+      # Update the hash of messages.
+      if (exists $dups->{$to_filter})
+	{
+	  ++$dups->{$to_filter};
+	  return 1;
+	}
+      else
+	{
+	  $dups->{$to_filter} = 0;
+	}
+    }
+  print $file $msg;
+  return 1;
+}
+
+
 # Store partial messages here. (See the 'partial' option.)
 use vars qw ($partial);
 $partial = '';
@@ -425,21 +511,21 @@ sub _print_message ($$%)
     }
 
   # Check for duplicate message if requested.
-  if ($opts{'uniq_part'} != UP_NONE)
+  my $to_filter;
+  if ($opts{'uniq_part'} ne UP_NONE)
     {
       # Which part of the error should we match?
-      my $to_filter;
-      if ($opts{'uniq_part'} == UP_TEXT)
+      if ($opts{'uniq_part'} eq UP_TEXT)
 	{
 	  $to_filter = $message;
 	}
-      elsif ($opts{'uniq_part'} == UP_LOC_TEXT)
+      elsif ($opts{'uniq_part'} eq UP_LOC_TEXT)
 	{
 	  $to_filter = $msg;
 	}
       else
 	{
-	  confess "unknown value for uniq_part: " . $opts{'uniq_part'};
+	  $to_filter = $opts{'uniq_part'};
 	}
 
       # Do we want local or global uniqueness?
@@ -469,13 +555,21 @@ sub _print_message ($$%)
 	}
     }
   my $file = $opts{'file'};
-  print $file $msg;
+  if ($opts{'ordered'} && $opts{'queue'})
+    {
+      _enqueue ($opts{'queue'}, $opts{'queue_key'}, $opts{'uniq_scope'},
+		$to_filter, $msg, $file);
+    }
+  else
+    {
+      print $file $msg;
+    }
   return 1;
 }
 
 =item C<msg ($channel, $location, $message, [%options])>
 
-Emit a message on C<$channel>, overriding some options of the channel  with
+Emit a message on C<$channel>, overriding some options of the channel with
 those specified in C<%options>.  Obviously C<$channel> must have been
 registered with C<register_channel>.
 
@@ -493,10 +587,10 @@ the following would be output:
   foo.c:10: unused variable `mumble'
 
 C<$location> can also be an instance of C<Automake::Location>.  In this
-case the stack of contexts will be displayed in addition.
+case, the stack of contexts will be displayed in addition.
 
 If C<$message> contains newline characters, C<$location> is prepended
-to each line.  For instance
+to each line.  For instance,
 
   msg 'error', 'somewhere', "1st line\n2nd line";
 
@@ -519,7 +613,7 @@ both print
 =cut
 
 
-use vars qw (@backlog %buffering @chain);
+use vars qw (@backlog %buffering);
 
 # See buffer_messages() and flush_messages() below.
 %buffering = ();	# The map of channel types to buffer.
@@ -560,7 +654,12 @@ sub msg ($$;$%)
 
       # Die on fatal messages.
       confess if $opts{'backtrace'};
-      exit $exit_code if $opts{'type'} eq 'fatal';
+      if ($opts{'type'} eq 'fatal')
+        {
+	  # flush messages explicitly here, needed in worker threads.
+	  STDERR->flush;
+	  exit $exit_code;
+	}
     }
 }
 
@@ -643,8 +742,8 @@ list for later processing.
 
 This backlog of messages is processed when C<flush_messages> is
 called, with the current channel options (not the options in effect,
-at the time of C<msg>).  So for instance if some channel was silenced
-in the meantime, messages to this channels will not be print.
+at the time of C<msg>).  So for instance, if some channel was silenced
+in the meantime, messages to this channel will not be printed.
 
 C<flush_messages> cancels the effect of C<buffer_messages>.  Following
 calls to C<msg> are processed immediately as usual.
@@ -669,6 +768,33 @@ sub flush_messages ()
   @backlog = ();
 }
 
+=item C<setup_channel_queue ($queue, $key)>
+
+Set the queue to fill for each channel that is ordered,
+and the key to use for serialization.
+
+=cut
+sub setup_channel_queue ($$)
+{
+  my ($queue, $key) = @_;
+  foreach my $channel (keys %channels)
+    {
+      setup_channel $channel, queue => $queue, queue_key => $key
+        if $channels{$channel}{'ordered'};
+    }
+}
+
+=item C<pop_channel_queue ($queue)>
+
+pop a message off the $queue; the key has already been popped.
+
+=cut
+sub pop_channel_queue ($)
+{
+  my ($queue) = @_;
+  return _dequeue ($queue);
+}
+
 =back
 
 =head1 SEE ALSO
